(*
 * Doubly-linked list.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2003 Mojave Group, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)

(*
 * Type of elements in the list.
 *)
type 'a elem =
   { mutable elem_pred : 'a elem option ref;
     elem_succ : 'a elem option ref;
     elem_item : 'a
   }

(*
 * The head of the list.
 *)
type 'a t = 'a elem option ref

(*
 * Create a new empty list.
 *)
let create_list () =
   ref None

(*
 * Create a new element.
 *)
let create_elem item =
   { elem_pred = ref None;
     elem_succ = ref None;
     elem_item = item
   }

(*
 * Get the item.
 *)
let item elem =
   elem.elem_item

let pred elem =
   !(elem.elem_pred)

let succ elem =
   !(elem.elem_succ)

(*
 * Unlink from the list.
 *)
let unlink elem =
   (* Unlink the node from its current list *)
   let pred = elem.elem_pred in
   let succ = !(elem.elem_succ) in
      pred := succ;
      match succ with
         Some next ->
            next.elem_pred <- pred
       | None ->
            ()

(*
 * Insert to head of the list.
 *)
let insert list elem =
   let next = !list in
      (* Add to the list *)
      list := Some elem;

      (* Adjust links *)
      elem.elem_pred <- list;
      elem.elem_succ := next;
      match next with
         Some next ->
            next.elem_pred <- elem.elem_succ
       | None ->
            ()

(*
 * Iterate through the node list.
 *)
let iter list f =
   let rec iter elem_opt =
      match elem_opt with
         Some elem ->
            let next = !(elem.elem_succ) in
               f elem;
               iter next
       | None ->
            ()
   in
      iter !list

(*
 * Fold through the elem list.
 *)
let fold list f x =
   let rec fold x elem =
      match elem with
         Some elem ->
            let next = !(elem.elem_succ) in
            let x = f x elem in
               fold x next
       | None ->
            x
   in
      fold x !list

(*
 * Find a particular elem.
 *)
let find list f =
   let rec find elem =
      match elem with
         Some elem ->
               if f elem then
                  elem
               else
                  find (!(elem.elem_succ))
       | None ->
            raise Not_found
   in
      find !list

(*
 * Test for empty.
 *)
let is_empty list =
   !list = None

(*
 * Get the head of a list.
 *)
let head list =
   match !list with
      Some elem ->
         elem
    | None ->
         raise (Invalid_argument "head")

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
