(*
 * Command utilities.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2003 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)
open Lm_printf

open Fmarshal

open Omake_node
open Omake_marshal
open Omake_cache_type
open Omake_ir_print

(************************************************************************
 * Command utilities
 *)

(*
 * Create a command line from a command string.
 *   Flags:
 *      '@' : QuietFlag
 *      '-' : AllowFailureFlag
 *)
let command_flags line =
   let len = String.length line in
   let rec process flags i =
      if i = len then
         flags, ""
      else
         match line.[i] with
            '@' ->
               process (QuietFlag :: flags) (succ i)
          | '-' ->
               process (AllowFailureFlag :: flags) (succ i)
          | '+' ->
               process flags (succ i)
          | _ ->
               let line =
                  if i = 0 then
                     line
                  else
                     String.sub line i (len - i)
               in
                  flags, line
   in
      process [] 0

(*
 * Parse the command lines from the strings.
 *)
let rec parse_commands venv dir target loc lines =
   match lines with
      line :: lines ->
         (match line with
             CommandArgv [] ->
                parse_commands venv dir target loc lines
           | CommandArgv (exe :: args) ->
                let flags, exe = command_flags exe in
                   if exe = "" then
                      match args with
                         _ :: _ ->
                            let command =
                               { command_loc    = loc;
                                 command_dir    = dir;
                                 command_target = target;
                                 command_flags  = flags;
                                 command_venv   = venv;
                                 command_inst   = CommandArgv args
                               }
                            in
                               command :: parse_commands venv dir target loc lines
                       | [] ->
                            parse_commands venv dir target loc lines
                   else
                      let command =
                         { command_loc    = loc;
                           command_dir    = dir;
                           command_target = target;
                           command_flags  = flags;
                           command_venv   = venv;
                           command_inst   = CommandArgv (exe :: args)
                         }
                      in
                         command :: parse_commands venv dir target loc lines
           | CommandEval _ ->
                let command =
                   { command_loc    = loc;
                     command_dir    = Dir.cwd ();
                     command_target = target;
                     command_flags  = [];
                     command_venv   = venv;
                     command_inst   = line
                   }
                in
                   command :: parse_commands venv dir target loc lines)
    | [] ->
         []

(*
 * Allow output in the command.
 *)
let command_allow_output command =
   { command with command_flags = AllowOutputFlag :: command.command_flags }

(*
 * Equality ignores flags.
 * It is assumed that all evaluated commands are the same.
 * This is because the rule in which the command is used
 * will be out-of-date if any of the OMakefiles changes.
 *)
let command_line_equal line1 line2 =
   match line1.command_inst, line2.command_inst with
      CommandArgv s1, CommandArgv s2 ->
         s1 = s2
    | CommandEval _, CommandEval _ ->
         true
    | CommandArgv _, CommandEval _
    | CommandEval _, CommandArgv _ ->
         false

let commands_are_equal commands1 commands2 =
   let len1 = List.length commands1 in
   let len2 = List.length commands2 in
      len1 = len2 && List.for_all2 command_line_equal commands1 commands2

(*
 * Remove the environment from the commands.
 *)
let squash_command_inst command =
   match command with
      CommandEval _ ->
         CommandEval ()
    | CommandArgv sl ->
         CommandArgv sl

let squash_command_line line =
   { line with command_venv = ();
               command_inst = squash_command_inst line.command_inst
   }

let squash_command_lines commands =
   List.map squash_command_line commands

(*
 * Print a command.
 *)
let rec pp_print_argv buf argv =
   match argv with
      [arg] ->
         pp_print_string buf arg
    | arg :: argv ->
         pp_print_string buf arg;
         pp_print_char buf ' ';
         pp_print_argv buf argv
    | [] ->
         ()

let pp_print_command_inst buf inst =
   match inst with
      CommandArgv argv ->
         pp_print_argv buf argv
    | CommandEval _ ->
         pp_print_string buf "<exp>"

let pp_print_command_line buf line =
   pp_print_command_inst buf line.command_inst

let pp_print_command_lines buf lines =
   List.iter (fun line -> fprintf buf "@ %a" pp_print_command_line line) lines

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
